// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#ifndef ARROW_UTIL_VARIANT_H
#define ARROW_UTIL_VARIANT_H

#include "arrow/vendored/variant/variant.hpp"  // IWYU pragma: export
#include "arrow/vendored/variant/variant_io.hpp"

namespace arrow {
namespace util {

using mapbox::util::apply_visitor;  // seems akin to std::visit
using mapbox::util::bad_variant_access;
using mapbox::util::get;
using mapbox::util::variant;

}  // namespace util
}  // namespace arrow

#endif  // ARROW_UTIL_VARIANT_H
